"""
# Description

Common constants and default values used in the QRotor subpackage.

Bond lengths and angles were obtained from MAPbI3,
see [*Cryst. Growth Des.* 2024, 24, 391−404](https://doi.org/10.1021/acs.cgd.3c01112).

---
"""


import numpy as np
import aton.phys as phys


# Distance between Carbon and Hydrogen atoms (measured from MAPbI3)
distance_CH = 1.09285   # Angstroms
"""Distance of the C-H bond, in Angstroms."""
distance_NH = 1.040263  # Angstroms
"""Distance of the N-H bond, in Angstroms."""

# Angles between atoms:  C-C-H  or  N-C-H  etc (from MAPbI3)
angle_CH_external = 108.7223
"""External angle of the X-C-H bond, in degrees."""
angle_NH_external = 111.29016
"""External angle of the X-N-H bond, in degrees."""
angle_CH = 180 - angle_CH_external
"""Internal angle of the X-C-H bond, in degrees."""
angle_NH = 180 - angle_NH_external
"""Internal angle of the X-N-H bond, in degrees."""

# Rotation radius (calculated from distance and angle)
r_CH = distance_CH * np.sin(np.deg2rad(angle_CH)) * phys.A_to_m
"""Rotation radius of the methyl group, in meters."""
r_NH = distance_NH * np.sin(np.deg2rad(angle_NH)) * phys.A_to_m
"""Rotation radius of the amine group, in meters."""

# Inertia, SI units
I_CH = 3 * (phys.atoms['H'].mass * r_CH**2)
"""Inertia of CH3, in uma·m^2."""
I_CD = 3 * (phys.atoms['H'].isotope[2].mass * r_CH**2)
"""Inertia of CD3, in uma·m^2."""
I_NH = 3 * (phys.atoms['H'].mass * r_NH**2)
"""Inertia of NH3, in uma·m^2."""
I_ND = 3 * (phys.atoms['H'].isotope[2].mass * r_NH**2)
"""Inertia of ND3, in uma·m^2."""

# Rotational energy.
B_CH = ((phys.hbar_eV**2) / (2 * I_CH)) * phys.J_to_eV
"""Rotational energy of CH3, in eV·s/uma·m^2."""
B_CD = ((phys.hbar**2) / (2 * I_CD)) * phys.J_to_eV
"""Rotational energy of CD3, in eV·s/uma·m^2."""
B_NH = ((phys.hbar**2) / (2 * I_NH)) * phys.J_to_eV
"""Rotational energy of NH3, in eV·s/uma·m^2."""
B_ND = ((phys.hbar**2) / (2 * I_ND)) * phys.J_to_eV
"""Rotational energy of ND3, in eV·s/uma·m^2."""

# Potential constants from titov2023 [C1, C2, C3, C4, C5]
constants_titov2023 = [
    [2.7860, 0.0130,-1.5284,-0.0037,-1.2791],
    [2.6507, 0.0158,-1.4111,-0.0007,-1.2547],
    [2.1852, 0.0164,-1.0017, 0.0003,-1.2061],
    [5.9109, 0.0258,-7.0152,-0.0168, 1.0213],
    [1.4526, 0.0134,-0.3196, 0.0005,-1.1461]
    ]
"""Potential constants from titov2023."""
constants_titov2023_zero = [
    [0,0,0,0,0]
    ]
"""Zero potential constants."""




"""
# Description

This module contains common classes for QRotor calculations.
These classes can be loaded directly as `aton.qrotor.Class()`.


# Index

| | |
| --- | --- |
| `QSys` | Quantum system, with all the data for a single calculation |
| `QExp` | Quantum experiment, contains several `QSys` objects |

---
"""


import numpy as np
from copy import deepcopy
from .constants import *
from aton.st import alias
from aton.spx import Plotting
from aton._version import __version__


class QSys:
    """Quantum system, contains all the data for a single calculation, with both inputs and outputs."""
    def __init__(
            self,
            comment: str = None,
            group: str = 'CH3',
            E_levels: int = 5,
            units = None,               ################ TODO CHECK THAT THIS WORKS. previously = []
            correct_potential_offset: bool = True,
            save_eigenvectors: bool = False,
            gridsize: int = None,
            grid = None,
            potential_name: str = None,
            potential_constants: list = None,
            potential_values = None,
            ):
        ## Technical
        self.comment: str = comment
        """Custom comment for the dataset."""
        self.group: str = group
        """Chemical group, methyl or amine: `'CH3'`, `'CD3'`, `'NH3'`, `'ND3'`."""
        self.set_group(group)  # Normalise the group name, and set the value of B
        self.E_levels: int = E_levels
        """Number of energy levels to be studied."""
        self.units = units
        """List containing the units in use, e.g. ['meV']."""     #############  TODO remove the need for a list
        self.correct_potential_offset: bool = correct_potential_offset
        """Correct the potential offset as `V - min(V)` or not."""
        self.save_eigenvectors: bool = save_eigenvectors
        """Save or not the eigenvectors. Final file size will be bigger."""
        ## Potential
        self.gridsize: int = gridsize
        """Number of points in the grid."""
        self.grid = grid
        """The grid with the points to be used in the calculation. Can be set automatically over $2 \\Pi$ with `QSys.set_grid()`."""
        if not B:
            B = self.B
        self.B: float = B
        """Rotational inertia, as in $B=\\frac{\\hbar^2}{2I}."""
        self.potential_name: str = potential_name
        """Name of the desired potential: `'zero'`, `'titov2023'`, `'test'`...
        If empty or unrecognised, the custom potential values inside `QSys.potential_values` will be used. 
        """
        self.potential_constants: list = potential_constants
        """List of constants to be used in the calculation of the potential energy, in the `aton.qrotor.potential` module."""
        self.potential_values = potential_values
        """Numpy array with the potential values for each point in the grid.
        Can be calculated with a function available in the `qrotor.potential` module,
        or loaded externally with the `qrotor.potential.load()` function.
        """
        self.potential_offset: float = None
        """`min(V)` before offset correction when `QSys.correct_potential_offset = True`"""
        self.potential_min: float = None
        """`min(V)`"""
        self.potential_max: float = None
        """`max(V)`"""
        self.potential_max_B: float = None
        """Reduced `potential_max`, in units of B."""
        # Energies
        self.eigenvalues = None
        """Calculated eigenvalues of the system."""
        self.eigenvalues_B = None
        """Reduced `eigenvalues`, in units of B."""
        self.eigenvectors = None
        """Eigenvectors, if `save_eigenvectors` is True. Beware of the file size."""
        self.energy_barrier: float = None
        """`max(V) - min(eigenvalues)`"""
        self.first_transition: float = None
        """eigenvalues[1] - eigenvalues[0]"""
        self.runtime: float = None
        """Time taken to solve the eigenvalues."""

    def summary(self):
        return {
            'comment': self.comment,
            'runtime': self.runtime,
            'group': self.gropu,
            'gridsize': self.gridsize,
            'B': self.B,
            'potential_name': self.potential_name,
            'potential_constants': self.potential_constants.tolist() if isinstance(self.potential_constants, np.ndarray) else self.potential_constants,
            'potential_offset': self.corrected_potential_offset,
            'potential_min': self.potential_min,
            'potential_max': self.potential_max,
            'potential_max / B': self.potential_max_B,
            'eigenvalues': self.eigenvalues.tolist() if isinstance(self.eigenvalues, np.ndarray) else self.eigenvalues,
            'eigenvalues / B': self.eigenvalues_B.tolist() if isinstance(self.eigenvalues_B, np.ndarray) else self.eigenvalues_B,
            'energy_barrier': self.energy_barrier,
            'first_transition': self.first_transition,
        }

    def set_grid(self, gridsize:int=None):
        """Sets the `QSys.grid` to the specified `gridsize`."""
        if gridsize is not None:
            self.gridsize = gridsize
        if self.gridsize is None:
            raise ValueError('QSys.gridsize is not set yet!.')
        self.grid = np.linspace(0, 2*np.pi, self.gridsize)
        return self
    
    def set_group(self, group:str=None, B:float=None):
        """Normalise `QSys.group` name, and set `QSys.B` based on it."""
        for name in alias.chemical['CH3']:
            if group.lower() == name:
                self.group = 'CH3'
                if not B:
                    B = B_CH
                self.B = B
                return self
        for name in alias.chemical['CD3']:
            if group.lower() == name:
                self.group = 'CD3'
                if not B:
                    B = B_CD
                self.B = B
                return self
        for name in alias.chemical['NH3']:
            if group.lower() == name:
                self.group = 'NH3'
                if not B:
                    B = B_NH
                self.B = B
                return self
        for name in alias.chemical['ND3']:
            if group.lower() == name:
                self.group = 'ND3'
                if not B:
                    B = B_ND
                self.B = B
                return self
        self.group = group  # No match was found
        return self


class QExp:
    def __init__(self,
                 comment: str = None,
                 systems: list = [],
                 plotting: Plotting = None,
                 ):
        self.version = __version__
        """Version of the package used to generate the data."""
        self.comment: str = comment
        """Custom comment for the dataset."""
        if isinstance(systems, QSys):
            systems = [systems]
        self.systems = systems
        """List containing the calculated `QSys` objects."""
        self.plotting: Plotting = plotting
        """`aton.spx.classes.Plotting` object (not implemented)."""

    def add(self, *args):
        """Adds more systems to `self.systems` from the introduced `QSys` or `QExp` objects."""
        for value in args:
            if isinstance(value, QExp):
                self.systems.extend(value.systems)
                self.version = value.version if len(self.systems) == 0 else self.version
                self.comment = value.comment if self.comment is None else self.comment
                self.plotting = value.plotting if self.plotting is None else self.plotting
            elif isinstance(value, QSys):
                self.systems.append(value)
            else:
                raise TypeError(f'QExp.add() can only add QExp and/or QSys objects, not {type(value)}.')

    def get_energies(self):
        """Returns a list with all `QSys.eigenvalues` values."""
        energies = []
        for i in self.systems:
            if all(i.eigenvalues):
                energies.append(i.eigenvalues)
            else:
                energies.append(None)
        return energies

    def get_gridsizes(self):
        """Returns a list with all `QSys.gridsize` values."""
        gridsizes = []
        for i in self.systems:
            if i.gridsize:
                gridsizes.append(i.gridsize)
            else:
                gridsizes.append(None)
        return gridsizes

    def get_runtimes(self):
        """Returns a list with all `QSys.runtime` values."""
        runtimes = []
        for i in self.systems:
            if i.runtime:
                runtimes.append(i.runtime)
            else:
                runtimes.append(None)
        return runtimes

    def get_gropus(self):
        """Returns a list with all `QSys.group` values."""
        groups = []
        for i in self.systems:
            if i.group not in groups:
                groups.append(i.group)
        return groups

    def sort_by_potential_values(self):
        """Returns the `QExp` object, sorted by `QSys.potential_values`."""
        grouped_data = self.group_by_potential_values()
        data = QExp(
            version = self.version,
            comment = self.comment,
            plotting = self.plotting,
        )
        for dataset in grouped_data:
            data.add(dataset)
        return data

    def group_by_potential_values(self):
        """Returns an array of grouped `QExp` objects with the same `QSys.potential_values`."""
        print('Grouping Experiment by potential_values...')
        grouped_data = []
        for system in self.systems:
            data = QExp()
            data.comment = self.comment
            data.systems.append(system)
            new_group = True
            for data_i in grouped_data:
                if np.array_equal(system.potential_values, data_i.systems[0].potential_values):
                    data_i.systems.append(system)
                    new_group = False
                    break
            if new_group:
                print('New potential_values found')
                grouped_data.append(data)
        return grouped_data

    def sort_by_gridsize(self):
        """Returns the same `QExp`, sorted by `QSys.gridsize`."""
        self.systems = sorted(self.systems, key=lambda sys: sys.gridsize)
        return self

    def get_ideal_E(self, E_level):
        """Calculates the ideal energy for a specified `E_level` for a convergence test. Only for 'zero' potential."""
        real_E_level = None
        if self.systems[0].potential_name == 'zero':
            if E_level % 2 == 0:
                real_E_level = E_level / 2
            else:
                real_E_level = (E_level + 1) / 2
            ideal_E = int(real_E_level ** 2)
            return ideal_E
        else:
            print("WARNING:  get_ideal_E() only valid for potential_name='zero'")
            return
    
    def discard_shit(self):
        """Discard data that takes too much space, like eigenvectors, potential values and grids."""
        for dataset in self.systems:
            dataset.eigenvectors = None
            dataset.potential_values = None
            dataset.grid = None
        return self




"""
# *Ab-initio* interfaces

This module contains interfaces for several *ab-initio* calculation softwares.
These interfaces can be easily expanded with the `aton.txt` module.


# Index

| | |  
| --- | --- |  
| `aton.interface.qe`      | Interface for [Quantum ESPRESSO](https://www.quantum-espresso.org/)'s [pw.x](https://www.quantum-espresso.org/Doc/INPUT_PW.html) module |  
| `aton.interface.phonopy` | Interface for [Phonopy](https://phonopy.github.io/phonopy/) calculations |  
| `aton.interface.castep`  | Interface for [CASTEP](https://castep-docs.github.io/castep-docs/) calculations |  
| `aton.interface.slurm` | Batch jobs via [Slurm](https://slurm.schedmd.com/) |


# Examples

## Quantum ESPRESSO

To read the output from a Quantum ESPRESSO pw.x calculation,  
```python
from aton.interface import qe
# Read to a dictionary
calculation = qe.read_out('relax.out')
calculation.keys()  # See the available values
# Final energy from the calculation
energy = calculation['Energy']
```

To modify values from an input file,  
```python
from aton.interface import qe
# Add a hydrogen atom to a specific position
qe.add_atom('H  0.10  0.20  0.30')
# Set the input ecutwfc value
qe.set_value('relax.in', 'ecutwfc', 60.0)
```

Check the full `aton.interface.qe` API reference for more details.


## Phonopy

To perform a phonon calculation from a relaxed structure via Quantum ESPRESSO,  
```python
from aton import interface
# Create the supercell inputs
interface.phonopy.make_supercells(dimension='2 2 2')
# Sbatch to a cluster
interface.slurm.sbatch('supercell-', 'template.slurm')
```

Check the full `aton.interface.phonopy` API reference for more details.


## CASTEP

To read output values from a CASTEP calculation,  
```python
from aton.interface import castep
# Read the output
output = castep.read_castep('calculation.castep')
# Get the final energy
energy = output['Energy']
```

Check the full `aton.interface.castep` API reference for more details.

"""


from . import qe
from . import phonopy
from . import castep
from . import slurm




"""
# Description

Functions to work with [Phonopy](https://phonopy.github.io/phonopy/) calculations,
along with [Quantum ESPRESSO](https://www.quantum-espresso.org/).  


# Index

| | |
| --- | --- |
| `make_supercells()` | Build supercell SCF inputs for phonon calculations |


# Examples

To create the supercells and run the phonon calculations
from a folder with `relax.in` and `relax.out` files,
using a `template.slurm` file,
```python
from aton import interface
interface.phonopy.make_supercells()
interface.slurm.sbatch('supercell-', 'scf.slurm')
```

---
"""


import os
import re
from aton._version import __version__
import aton.st.file as file
import aton.st.call as call
import aton.txt.find as find
import aton.txt.edit as edit # text
import aton.txt.extract as extract
import aton.interface.qe as qe
import aton.interface.slurm as slurm


def make_supercells(
        dimension:str='2 2 2',
        relax_in:str='relax.in',
        relax_out:str='relax.out',
        folder:str=None,
        slurm_template:str=None,
    ) -> None:
    """
    Creates the supercell inputs of a given `dimension` ('2 2 2' by default),
    from the `relax_in` and `relax_out` files in the `folder`
    ('relax.in', 'relax.out' and CWD by default, respectively),
    needed for the Phonopy calculations with Quantum ESPRESSO.

    If `slurm_template` is present,
    it checks it with `aton.interface.slurm.check_template()`.
    """
    print(f'\nWelcome to thotpy.phonopy {__version__}\n'
          'Creating all supercell inputs with Phonopy for Quantum ESPRESSO...\n')
    qe.scf_from_relax(folder, relax_in, relax_out)
    _supercells_from_scf(dimension, folder)
    _copy_scf_header_to_supercells(folder)
    print('\n------------------------------------------------------\n'
          'PLEASE CHECH BELOW THE CONTENT OF THE supercell-001.in\n'
          '------------------------------------------------------\n')
    call.bash('head -n 100 supercell-001.in')
    print('\n------------------------------------------------------\n'
          'PLEASE CHECH THE CONTENT OF THE supercell-001.in\n'
          'The first 100 lines of the input were printed above!\n'
          '------------------------------------------------------\n\n'
          'If it seems correct, run the calculations with thotpy.phonopy.sbatch()\n')
    if slurm_template:
        slurm.check_template(slurm_template, folder)
    return None


def _supercells_from_scf(
        dimension:str='2 2 2',
        folder:str=None,
        scf:str='scf.in'
    ) -> None:
    """
    Creates supercells of a given `dimension` (`2 2 2` by default) inside a `folder`,
    from a Quantum ESPRESSO `scf` input (`scf.in` by default).
    """
    print(f'\naton.interface.phonopy {__version__}\n')
    folder = call.here(folder)
    scf_in = file.get(folder, scf, True)
    if scf_in is None:
        raise ValueError('No SCF input found in path!')
    call.bash(f'phonopy --qe -d --dim="{dimension}" -c {scf_in}')
    return None


def _copy_scf_header_to_supercells(
        folder:str=None,
        scf:str='scf.in',
    ) -> None:
    """Paste the header from the `scf` file in `folder` to the supercells created by Phonopy."""
    print(f'\naton.interface.phonopy {__version__}\n'
          f'Adding headers to Phonopy supercells for Quantum ESPRESSO...\n')
    folder = call.here(folder)
    # Check if the header file, the scf.in, exists
    scf_file = file.get(folder, scf, True)
    if scf_file is None:
        raise ValueError('No header file found in path!')
    # Check if the supercells exist
    supercells = file.get_list(folder, 'supercell-')
    if supercells is None:
        raise ValueError('No supercells found in path!')
    # Check if the supercells contains '&CONTROL' and abort if so
    supercell_sample = supercells[0]
    is_control = find.lines(supercell_sample, r'(&CONTROL|&control)', 1, 0, False, True)
    if is_control:
        raise ValueError('Supercells already contain &CONTROL! Did you do this already?')
    # Check if the keyword is in the scf file
    is_header = find.lines(scf_file, r'ATOMIC_SPECIES', 1, 0, False, False)
    if not is_header:
        raise ValueError('No ATOMIC_SPECIES found in header!')
    # Copy the scf to a temp file
    temp_scf = '_scf_temp.in'
    file.copy(scf_file, temp_scf)
    # Remove the top content from the temp file
    edit.delete_under(temp_scf, 'K_POINTS', -1, 2, False)
    # Find the new number of atoms and replace the line
    updated_values = find.lines(supercell_sample, 'ibrav', 1)  # !    ibrav = 0, nat = 384, ntyp = 5
    if not updated_values:
        print("!!! Okay listen, this is weird. This code should never be running, "
              "but for some reson we couldn't find the updated values in the supercells. "
              "Please, introduce the NEW NUMBER OF ATOMS in the supercells manually (int):")
        nat = int(input('nat = '))
    else:
        nat = extract.number(updated_values[0], 'nat')
    qe.set_value(temp_scf, 'nat', nat)
    # Remove the lattice parameters, since Phonopy already indicates units
    qe.set_value(temp_scf, 'celldm(1)', '')
    qe.set_value(temp_scf, 'A', '')
    qe.set_value(temp_scf, 'B', '')
    qe.set_value(temp_scf, 'C', '')
    qe.set_value(temp_scf, 'cosAB', '')
    qe.set_value(temp_scf, 'cosAC', '')
    qe.set_value(temp_scf, 'cosBC', '')
    # Add the header to the supercells
    with open(temp_scf, 'r') as f:
        header = f.read()
    for supercell in supercells:
        edit.insert_at(supercell, header, 0)
    # Remove the temp file
    os.remove('_scf_temp.in')
    print('Done!')
    return None

